# Generated by Django 2.2.10 on 2020-07-24 08:09

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("partner", "0005_auto_20181115_1953"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="stockrecord",
            name="cost_price",
        ),
        migrations.RemoveField(
            model_name="stockrecord",
            name="price_retail",
        ),
        migrations.AlterField(
            model_name="stockrecord",
            name="price_excl_tax",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                max_digits=12,
                null=True,
                verbose_name="Price",
            ),
        ),
        migrations.RenameField(
            model_name="stockrecord",
            old_name="price_excl_tax",
            new_name="price",
        ),
    ]



# Generated by Django 2.2.13 on 2021-03-04 12:01

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("partner", "0006_auto_20200724_0909"),
        ("partner", "0006_auto_20190612_0844"),
    ]

    operations = []



from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class, get_model

ReportGenerator = get_class("dashboard.reports.reports", "ReportGenerator")
ReportCSVFormatter = get_class("dashboard.reports.reports", "ReportCSVFormatter")
ReportHTMLFormatter = get_class("dashboard.reports.reports", "ReportHTMLFormatter")

StockAlert = get_model("partner", "StockAlert")


class StockAlertReportCSVFormatter(ReportCSVFormatter):
    filename_template = "stock_alerts.csv"

    def generate_csv(self, response, stockalerts):
        writer = self.get_csv_writer(response)
        header_row = [
            _("Product"),
            _("Partner"),
            _("Low stock threshold"),
            _("Current available stock"),
            _("Date alert raised"),
        ]
        writer.writerow(header_row)

        for alert in stockalerts:
            row = [
                alert.stockrecord.product.get_title(),
                alert.stockrecord.partner.display_name,
                alert.stockrecord.low_stock_threshold,
                alert.stockrecord.net_stock_level,
                alert.date_created,
            ]
            writer.writerow(row)


class StockAlertReportHTMLFormatter(ReportHTMLFormatter):
    filename_template = "dashboard/reports/partials/stockalerts_report.html"


class StockAlertReportGenerator(ReportGenerator):
    code = "stock_alerts"
    description = _("Open stock alerts")
    date_range_field_name = "date_created"

    formatters = {
        "CSV_formatter": StockAlertReportCSVFormatter,
        "HTML_formatter": StockAlertReportHTMLFormatter,
    }

    def generate(self):
        # Only open stock alerts are usefull, as closed ones are already handled.
        stock_alerts = self.get_queryset()
        return self.formatter.generate_response(stock_alerts)

    def get_queryset(self):
        return StockAlert.objects.filter(status=StockAlert.OPEN)



# -*- coding: utf-8 -*-
import logging

from oscar.apps.partner.strategy import Structured
from oscar.core.loading import get_classes

from ocyan.core.fender import config
from ocyan.core.utils import RequestCache

(
    ProductSpecifiedTaxInclusive,
    ProductSpecifiedTaxExclusive,
    UseCheapestChildProductStockRecord,
    StockRequired,
) = get_classes(
    "partner.mixins",
    [
        "ProductSpecifiedTaxInclusive",
        "ProductSpecifiedTaxExclusive",
        "UseCheapestChildProductStockRecord",
        "StockRequired",
    ],
)
logger = logging.getLogger(__name__)


class Selector(object):
    """
    Responsible for returning the appropriate strategy class for a given
    user/session.

    This can be called in three ways:

    #) Passing a request and user.  This is for determining
       prices/availability for a normal user browsing the site.

    #) Passing just the user.  This is for offline processes that don't
       have a request instance but do know which user to determine prices for.

    #) Passing nothing.  This is for offline processes that don't
       correspond to a specific user.  Eg, determining a price to store in
       a search index.

    """

    # pylint: disable=unused-argument
    def strategy(self, request=None, user=None, **kwargs):
        """
        Return an instanticated strategy instance
        """
        use_price_incl_tax = config.get("oscar", "use_price_incl_tax", True)
        if use_price_incl_tax:
            return NLInclusive(request)

        return NLExclusive(request)


# pylint: disable=abstract-method
class OncePerRequest(Structured):
    @RequestCache.once_per_request
    def fetch_for_product(self, product, stockrecord=None):
        return super(OncePerRequest, self).fetch_for_product(product, stockrecord)

    @RequestCache.once_per_request
    def fetch_for_parent(self, product):
        return super(OncePerRequest, self).fetch_for_parent(product)


class NLInclusive(
    UseCheapestChildProductStockRecord,
    StockRequired,
    ProductSpecifiedTaxInclusive,
    OncePerRequest,
):
    pass


class NLExclusive(
    UseCheapestChildProductStockRecord,
    StockRequired,
    ProductSpecifiedTaxExclusive,
    OncePerRequest,
):
    pass



from datetime import datetime, timedelta
from decimal import Decimal as D
from unittest import mock

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.test import TestCase
from django.test.client import RequestFactory
from freezegun import freeze_time

from oscar.core.loading import get_class, get_model

import ocyan.core.fender

from ocyan.plugin.oscar_partner.partner.constants import LOCAL_VAT_APPLICABLE_KEY

Product = get_model("catalogue", "Product")
ProductClass = get_model("catalogue", "ProductClass")
Basket = get_model("basket", "Basket")
Order = get_model("order", "Order")
LocalEUVatSettings = get_model("oscar_partner", "LocalEUVatSettings")

Selector = get_class("partner.strategy", "Selector")

User = get_user_model()

EXCL_TAX_SETTINGS = {"oscar": {"use_price_incl_tax": False}}


class BaseLocalVatTest(TestCase):
    fixtures = ["partner/auth.json", "partner/catalogue.json"]

    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.get(username="henk")
        self.client.force_login(self.user)
        self.COUNTRY_OF_ORIGIN = settings.OSCAR_VAT_COUNTRY_OF_ORIGIN_CODE
        self.product = Product.objects.get(pk=1)
        self.eu_settings = LocalEUVatSettings.load()

    def treshold_exceeded(self):
        self.eu_settings.vat_threshold_exceeded = True
        self.eu_settings.save()
        cache.delete(LOCAL_VAT_APPLICABLE_KEY)

    def treshold_un_exceeded(self):
        self.eu_settings.vat_threshold_exceeded = False
        self.eu_settings.save()
        cache.delete(LOCAL_VAT_APPLICABLE_KEY)


class LocalVatExclTaxTest(BaseLocalVatTest):
    @mock.patch.object(ocyan.core.fender.config, "settings", EXCL_TAX_SETTINGS)
    def test_country_of_origin_without_treshold_exceeded(self):
        request = self.factory.get("/")
        request.META["shipping_country"] = self.COUNTRY_OF_ORIGIN

        strategy = Selector().strategy()
        strategy.user = self.user
        strategy.request = request
        purchase_info = strategy.fetch_for_product(self.product)

        self.assertEqual(purchase_info.price.excl_tax, 100)
        self.assertEqual(purchase_info.price.incl_tax, 121)

    @mock.patch.object(ocyan.core.fender.config, "settings", EXCL_TAX_SETTINGS)
    def test_country_of_origin_with_treshold_exceeded(self):
        self.treshold_exceeded()
        request = self.factory.get("/")
        request.META["shipping_country"] = self.COUNTRY_OF_ORIGIN

        strategy = Selector().strategy()
        strategy.user = self.user
        strategy.request = request
        purchase_info = strategy.fetch_for_product(self.product)

        self.assertEqual(purchase_info.price.excl_tax, 100)
        self.assertEqual(purchase_info.price.incl_tax, 121)

    @mock.patch.object(ocyan.core.fender.config, "settings", EXCL_TAX_SETTINGS)
    def test_dk_with_treshold_exceeded(self):
        self.treshold_exceeded()
        request = self.factory.get("/")
        request.META["shipping_country"] = "DK"

        strategy = Selector().strategy()
        strategy.user = self.user
        strategy.request = request
        purchase_info = strategy.fetch_for_product(self.product)

        self.assertEqual(purchase_info.price.excl_tax, 100)
        self.assertEqual(purchase_info.price.incl_tax, 125)

    @mock.patch.object(ocyan.core.fender.config, "settings", EXCL_TAX_SETTINGS)
    def test_dk_without_treshold_exceeded(self):
        self.treshold_un_exceeded()
        request = self.factory.get("/")
        request.META["shipping_country"] = "DK"

        strategy = Selector().strategy()
        strategy.user = self.user
        strategy.request = request
        purchase_info = strategy.fetch_for_product(self.product)

        self.assertEqual(purchase_info.price.excl_tax, 100)
        self.assertEqual(purchase_info.price.incl_tax, 121)


class LocalVatInclTaxTest(BaseLocalVatTest):
    def test_country_of_origin_without_treshold_exceeded(self):
        request = self.factory.get("/")
        request.META["shipping_country"] = self.COUNTRY_OF_ORIGIN

        strategy = Selector().strategy()
        strategy.user = self.user
        strategy.request = request
        purchase_info = strategy.fetch_for_product(self.product)

        self.assertEqual(purchase_info.price.excl_tax, D("82.6446"))
        self.assertEqual(purchase_info.price.incl_tax, 100)

    def test_country_of_origin_with_treshold_exceeded(self):
        self.treshold_exceeded()
        request = self.factory.get("/")
        request.META["shipping_country"] = self.COUNTRY_OF_ORIGIN

        strategy = Selector().strategy()
        strategy.user = self.user
        strategy.request = request
        purchase_info = strategy.fetch_for_product(self.product)

        self.assertEqual(purchase_info.price.excl_tax, D("82.6446"))
        self.assertEqual(purchase_info.price.incl_tax, 100)

    def test_dk_with_treshold_exceeded(self):
        self.treshold_exceeded()
        request = self.factory.get("/")
        request.META["shipping_country"] = "DK"

        strategy = Selector().strategy()
        strategy.user = self.user
        strategy.request = request
        purchase_info = strategy.fetch_for_product(self.product)

        self.assertEqual(purchase_info.price.excl_tax, D("82.6446"))
        self.assertEqual(purchase_info.price.incl_tax, D("103.3058"))

    def test_dk_without_treshold_exceeded(self):
        self.treshold_un_exceeded()
        request = self.factory.get("/")
        request.META["shipping_country"] = "DK"

        strategy = Selector().strategy()
        strategy.user = self.user
        strategy.request = request
        purchase_info = strategy.fetch_for_product(self.product)

        self.assertEqual(purchase_info.price.excl_tax, D("82.6446"))
        self.assertEqual(purchase_info.price.incl_tax, 100)


@freeze_time("2021-12-14 17:23:26")
class LocalVatSetterTest(BaseLocalVatTest):
    fixtures = ["partner/orders.json"] + BaseLocalVatTest.fixtures

    def test_treshold_is_set(self):
        yesterday = datetime.now() - timedelta(1)
        Order.objects.filter(pk=1).update(date_placed=yesterday)

        self.treshold_un_exceeded()
        self.assertEqual(self.eu_settings.vat_threshold_exceeded, False)

        request = self.factory.get("/")
        request.META["shipping_country"] = "FR"

        strategy = Selector().strategy()
        strategy.user = self.user
        strategy.request = request
        strategy.fetch_for_product(self.product)

        self.assertTrue(cache.get(LOCAL_VAT_APPLICABLE_KEY))

        self.eu_settings.refresh_from_db()
        self.assertEqual(self.eu_settings.vat_threshold_exceeded, False)



from django.conf import settings

from oscar.core.loading import get_class, get_classes

from ocyan.core.utils import round_to_four_dec
from ocyan.plugin.oscar.fender import InvalidTaxRateException

ExcludingTaxPrice, IncludingTaxPrice = get_classes(
    "partner.prices", ["ExcludingTaxPrice", "IncludingTaxPrice"]
)
(
    tax_from_price_incl_tax,
    get_vat_country_from_request,
    is_local_vat_applicable,
    get_vat_info_for_country,
) = get_classes(
    "partner.utils",
    [
        "tax_from_price_incl_tax",
        "get_vat_country_from_request",
        "is_local_vat_applicable",
        "get_vat_info_for_country",
    ],
)

COUNTRY_TAX_RATES = get_class("taxes.constants", "COUNTRY_TAX_RATES")
exclusive_tax_component = get_class("taxes.utils", "exclusive_tax_component")
tax_rate_as_fraction = get_class("taxes.taxrate", "tax_rate_as_fraction")


class IncludingTaxPriceSelectorFunctor:
    def __call__(
        self, user, request, vat_rate=None, is_shipping_charge=False, **kwargs
    ):
        if vat_rate is None:
            try:
                return IncludingTaxPrice(**kwargs)
            except TypeError as e:
                raise InvalidTaxRateException(str(e)) from e

        return self.get_price(
            incl_tax=kwargs["incl_tax"],
            currency=kwargs["currency"],
            vat_rate=vat_rate,
            vat_country=get_vat_country_from_request(request),
            is_shipping_charge=is_shipping_charge,
            basket=getattr(request, "basket", None),
            user=user,
        )

    def get_price(
        self,
        incl_tax,
        currency,
        vat_rate,
        vat_country,
        is_shipping_charge=False,
        basket=None,
        user=None,
    ):
        tax_percentage = get_vat_info_for_country(
            vat_rate,
            vat_country,
            is_shipping_charge,
            basket=basket,
            user=user,
        ).percentage

        # When the local vat is applicable and the relative price should be used we want to calculate the excl_tax price for the country of origin.
        # With the relative excl_tax we can calculate the correct tax amount for the designated shipping country.
        if (
            settings.OCYAN_PARTNER_RELATIVE_PRICE
            and is_local_vat_applicable(vat_country)
            and vat_country != settings.OSCAR_VAT_COUNTRY_OF_ORIGIN_CODE
        ):
            return self.get_relative_price_for_vat_country(
                incl_tax=incl_tax,
                currency=currency,
                vat_rate=vat_rate,
                tax_percentage=tax_percentage,
            )

        return IncludingTaxPrice(
            currency=currency,
            incl_tax=incl_tax,
            tax=tax_from_price_incl_tax(incl_tax, tax_percentage),
            tax_code=vat_rate,
        )

    def get_relative_price_for_vat_country(
        self, incl_tax, currency, vat_rate, tax_percentage
    ):
        COUNTRY_OF_ORIGIN_VAT_COMPONENT = exclusive_tax_component(
            tax_rate_as_fraction(vat_rate)
        )
        excl_tax = incl_tax / COUNTRY_OF_ORIGIN_VAT_COMPONENT
        return ExcludingTaxPrice(
            currency=currency,
            excl_tax=round_to_four_dec(excl_tax),
            tax=round_to_four_dec(excl_tax * tax_percentage),
            tax_code=vat_rate,
        )


IncludingTaxPriceSelector = IncludingTaxPriceSelectorFunctor()


class ExcludingTaxPriceSelectorFunctor:
    # pylint: disable=unused-argument
    def __call__(
        self, user, request, vat_rate=None, is_shipping_charge=False, **kwargs
    ):
        if vat_rate is None:
            try:
                return ExcludingTaxPrice(**kwargs)
            except TypeError as e:
                raise InvalidTaxRateException(str(e)) from e

        return self.get_price(
            excl_tax=kwargs["excl_tax"],
            currency=kwargs["currency"],
            vat_rate=vat_rate,
            vat_country=get_vat_country_from_request(request),
            is_shipping_charge=is_shipping_charge,
            basket=getattr(request, "basket", None),
            user=user,
        )

    def get_price(
        self,
        excl_tax,
        currency,
        vat_rate,
        vat_country,
        is_shipping_charge,
        basket=None,
        user=None,
    ):
        tax_percentage = get_vat_info_for_country(
            vat_rate,
            vat_country,
            is_shipping_charge,
            basket=basket,
            user=user,
        ).percentage
        return ExcludingTaxPrice(
            currency=currency,
            excl_tax=excl_tax,
            tax=round_to_four_dec(excl_tax * tax_percentage),
            tax_code=vat_rate,
        )


ExcludingTaxPriceSelector = ExcludingTaxPriceSelectorFunctor()



# Generated by Django 2.2.13 on 2021-07-16 12:38

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("wagtailcore", "0041_group_collection_permissions_verbose_name_plural"),
    ]

    operations = [
        migrations.CreateModel(
            name="LocalEUVatSettings",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("vat_threshold_exceeded", models.BooleanField(default=False)),
                (
                    "site",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="wagtailcore.Site",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.test import TestCase

from oscar.core.loading import get_class, get_model

Product = get_model("catalogue", "Product")
ProductClass = get_model("catalogue", "ProductClass")
Basket = get_model("basket", "Basket")
Selector = get_class("partner.strategy", "Selector")


class SelectCorrectStockRecordTest(TestCase):
    fixtures = ["taxrounding"]

    def test_price_for_standalone(self):
        product = Product.objects.get(pk=1)
        strategy = Selector().strategy()
        purchase_info = strategy.fetch_for_product(product)
        # we only want to know that there is a price available to make sure
        # the stockrecord selection is not broken
        self.assertTrue(purchase_info.price.excl_tax > 0)
        self.assertTrue(purchase_info.price.incl_tax > 0)

    def test_price_for_parent(self):
        product = Product.objects.get(pk=2)
        strategy = Selector().strategy()
        purchase_info = strategy.fetch_for_product(product)
        # parents don't have a price with fetch_for_product
        # as they don't have stockrecords
        self.assertIsNone(purchase_info.price.excl_tax)
        self.assertIsNone(purchase_info.price.incl_tax)

        # but fetch_for_parent should return a price based on the child
        # stockrecords
        purchase_info = strategy.fetch_for_parent(product)
        self.assertTrue(purchase_info.price.excl_tax > 0)
        self.assertTrue(purchase_info.price.incl_tax > 0)

    def test_price_for_child(self):
        product = Product.objects.get(pk=3)
        strategy = Selector().strategy()
        purchase_info = strategy.fetch_for_product(product)
        # we only want to know that there is a price available to make sure
        # the stockrecord selection is not broken
        self.assertTrue(purchase_info.price.excl_tax > 0)
        self.assertTrue(purchase_info.price.incl_tax > 0)

        # when no stock is tracked, we still should get a price
        ProductClass.objects.filter(id=322).update(track_stock=False)
        purchase_info = strategy.fetch_for_product(product)
        self.assertTrue(purchase_info.price.excl_tax > 0)
        self.assertTrue(purchase_info.price.incl_tax > 0)



# -*- coding: utf-8 -*-
from django.conf import settings

from oscar.apps.partner.prices import *  # pylint: disable=W0401,W0614


class TaxedPriceBase(Base):
    exists = is_tax_known = True

    @property
    def effective_price(self):
        if settings.OSCAR_OFFERS_INCL_TAX:
            return self.incl_tax
        else:
            return self.excl_tax


class ExcludingTaxPrice(TaxedPriceBase):
    """
    Specialised version of FixedPrice that must have tax passed.  It also
    specifies that offers should use the tax-inclusive price (which is the norm
    in the UK).
    """

    def __init__(self, currency, excl_tax, tax, tax_code=None):
        self.currency = currency
        self.excl_tax = excl_tax
        self.tax = tax
        self.tax_code = tax_code

    @property
    def incl_tax(self):
        if self.excl_tax is not None:
            return self.excl_tax + self.tax

        return None


class IncludingTaxPrice(TaxedPriceBase):
    def __init__(self, currency, incl_tax, tax, tax_code=None):
        self.currency = currency
        self.incl_tax = incl_tax
        self.tax = tax
        self.tax_code = tax_code

    @property
    def excl_tax(self):
        if self.incl_tax is not None:
            return self.incl_tax - self.tax

        return None



from oscar.apps.partner import availability as oscar


class StockRequired(oscar.StockRequired):
    # pylint: disable=unused-argument
    def __init__(self, product, stockrecord):
        super().__init__(stockrecord.net_stock_level)



from django.db import models
from django.db.models import functions as db_functions

from ocyan.plugin.oscar_partner.partner import mixins as ocyan


class UseCheapestChildProductStockRecord(ocyan.UseCheapestChildProductStockRecord):
    def get_product_stockrecords_queryset(self, product):
        return product.stockrecords.filter(
            price_currency__isnull=False, price__isnull=False
        )

    def get_product_children_queryset(self, product):
        return product.variants_ordered_by_price.filter(
            is_public=True,
            stockrecords__price_currency__isnull=False,
            stockrecords__price__isnull=False,
        )

    def select_stockrecord(self, product):
        pc = product.get_product_class()
        if pc.track_stock:
            return (
                self.get_product_stockrecords_queryset(product)
                .annotate(
                    db_effective_stock=db_functions.Coalesce(
                        "num_in_stock", models.Value(0)
                    )
                    - db_functions.Coalesce("num_allocated", models.Value(0))
                )
                .annotate(
                    has_stock=models.Case(
                        models.When(db_effective_stock__gt=0, then=models.Value(True)),
                        default=models.Value(False),
                        output_field=models.BooleanField(),
                    )
                )
                .order_by("-has_stock", "price", "-db_effective_stock")
            ).first()
        else:
            return self.get_product_stockrecords_queryset(product).first()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.core.utils


class Migration(migrations.Migration):
    dependencies = [
        ("partner", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="stockrecord",
            name="price_currency",
            field=models.CharField(
                default=oscar.core.utils.get_default_currency,
                max_length=12,
                verbose_name="Currency",
            ),
        ),
    ]



from ocyan.plugin.oscar_partner.partner import abstract_models as ocyan


# pylint: disable=attribute-defined-outside-init
class AbstractStockRecord(ocyan.AbstractStockRecord):
    @property
    def price_excl_tax(self):
        return self.price

    @price_excl_tax.setter
    def price_excl_tax(self, value):
        self.price = value

    @property
    def price_incl_tax(self):
        return self.price

    @price_incl_tax.setter
    def price_incl_tax(self, value):
        self.price = value

    class Meta(ocyan.AbstractStockRecord.Meta):
        abstract = True



from django.db import models

from wagtail.contrib.settings.models import BaseGenericSetting, register_setting


@register_setting
class LocalEUVatSettings(BaseGenericSetting):
    vat_threshold_exceeded = models.BooleanField(default=False)



# -*- coding: utf-8 -*-
import doctest

import ocyan.plugin.oscar_partner.partner.utils


def load_tests(loader, tests, ignore):  # pylint: disable=W0613
    tests.addTests(doctest.DocTestSuite(ocyan.plugin.oscar_partner.partner.utils))
    return tests


# Create your tests here.



from decimal import Decimal as D

from django.conf import settings
from stdnum.eu.vat import MEMBER_STATES

EU_LOCAL_VAT_TRESHOLD = D("10000")

EU_APPLICABLE_COUNTRIES = [
    country.upper()
    for country in MEMBER_STATES
    if country.upper() != settings.OSCAR_VAT_COUNTRY_OF_ORIGIN_CODE
]


LOCAL_VAT_APPLICABLE_KEY = "is_local_vat_applicable"



"""
Find stockrecords that have allocations that do not match the unshipped order lines
"""

from django.core.management.base import BaseCommand
from django.db import models
from django.db.models.functions import Coalesce

from oscar.core.loading import get_model

from ocyan.plugin.oscar_shipping.shipping.constants import SHIPPING_EVENT_SHIPPED

ShippingEventType = get_model("order", "ShippingEventType")
StockRecord = get_model("partner", "StockRecord")
Line = get_model("order", "Line")


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-dry-run", action="store_true", help="Really commit to the database"
        )
        parser.add_argument(
            "--ask", action="store_true", help="Ask for updating allocations"
        )
        parser.add_argument(
            "--shipping-event-code",
            default=SHIPPING_EVENT_SHIPPED,
            help="To check for a certain shipping event, specify the code",
        )

    def handle(self, *args, **options):
        et = ShippingEventType.objects.get(code=options["shipping_event_code"])

        lines_with_unshipped_items = Line.objects.filter(
            shipping_event_quantities__event__event_type=et,
            stockrecord=models.OuterRef("pk"),
        ).annotate(
            quantity_shipped=Coalesce(
                models.Sum("shipping_event_quantities__quantity"), 0
            ),
            total_quantity=models.Sum("quantity"),
            total_unshipped_quantity=models.F("total_quantity")
            - models.F("quantity_shipped"),
        )

        for sr in StockRecord.objects.annotate(
            total_unshipped_quantity=models.Subquery(
                lines_with_unshipped_items.values("total_unshipped_quantity")[:1]
            ),
        ).filter(~models.Q(total_unshipped_quantity=Coalesce("num_allocated", 0))):
            if sr.product.get_product_class().track_stock:
                self.stdout.write(
                    self.style.ERROR(
                        "Dit product heeft niet dezelfde aantal openstaande order items dan er allocated is %s, open: %i, allocated: %i"
                        % (sr, sr.total_unshipped_quantity, sr.num_allocated)
                    )
                )
                if options["no_dry_run"]:
                    sr.num_allocated = sr.total_unshipped_quantity
                    sr.save()
                elif options["ask"]:
                    result = input("Wil je je de allocated updaten? (y/n) ")
                    if result.lower() == "y":
                        sr.num_allocated = sr.total_unshipped_quantity
                        sr.save()



# Generated by Django 2.0.13 on 2019-06-12 08:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("partner", "0005_auto_20181115_1953"),
    ]

    operations = [
        migrations.AlterField(
            model_name="stockrecord",
            name="num_in_stock",
            field=models.IntegerField(
                blank=True, null=True, verbose_name="Number in stock"
            ),
        ),
    ]



# -*- coding: utf-8 -*-
import logging

from django.db import models
from django.db.models import functions as db_functions

from oscar.apps.partner.strategy import StockRequired as BaseStockRequired
from oscar.apps.partner.strategy import UseFirstStockRecord
from oscar.core.loading import get_class, get_classes

from ocyan.plugin.oscar.fender import InvalidTaxRateException

UnavailablePrice = get_class("partner.prices", "Unavailable")

ExcludingTaxPriceSelector, IncludingTaxPriceSelector = get_classes(
    "partner.taxbase", ["ExcludingTaxPriceSelector", "IncludingTaxPriceSelector"]
)

Unavailable = get_class("partner.availability", "Unavailable")
Available = get_class("partner.availability", "Available")
StockRequiredAvailability = get_class("partner.availability", "StockRequired")

logger = logging.getLogger(__name__)


class ProductSpecifiedTaxInclusive(object):
    """
    Pricing policiy where the price is declared on the stockrecord including
    tax and the tax rate is declared on the product.
    """

    def pricing_policy(self, product, stockrecord):
        if not stockrecord or stockrecord.price_excl_tax is None:
            return UnavailablePrice()

        try:
            return IncludingTaxPriceSelector(
                self.user,  # pylint: disable=no-member
                self.request,  # pylint: disable=no-member
                currency=stockrecord.price_currency,
                incl_tax=stockrecord.price_incl_tax,
                vat_rate=product.vat_rate,
            )
        except InvalidTaxRateException:
            logger.error("Tax missing for %s", product)
            return UnavailablePrice()

    def parent_pricing_policy(self, product, children_stock):
        stockrecords = [x[1] for x in children_stock if x[1] is not None]
        if not stockrecords:
            return UnavailablePrice()

        try:
            # We take price from first record
            stockrecord = stockrecords[0]

            return IncludingTaxPriceSelector(
                self.user,  # pylint: disable=no-member
                self.request,  # pylint: disable=no-member
                currency=stockrecord.price_currency,
                incl_tax=stockrecord.price_incl_tax,
                vat_rate=product.vat_rate,
            )
        except InvalidTaxRateException:
            logger.error("Tax missing for %s", product)
            return UnavailablePrice()


class ProductSpecifiedTaxExclusive(object):
    """
    Pricing policiy where the price is declared on the stockrecord excluding
    tax and the tax rate is declared on the product.
    """

    def pricing_policy(self, product, stockrecord):
        if not stockrecord or stockrecord.price_excl_tax is None:
            return UnavailablePrice()

        try:
            return ExcludingTaxPriceSelector(
                self.user,  # pylint: disable=no-member
                self.request,  # pylint: disable=no-member
                currency=stockrecord.price_currency,
                excl_tax=stockrecord.price_excl_tax,
                vat_rate=product.vat_rate,
            )
        except InvalidTaxRateException:
            logger.error("Tax missing for %s", product)
            return UnavailablePrice()

    def parent_pricing_policy(self, product, children_stock):
        stockrecords = [x[1] for x in children_stock if x[1] is not None]
        if not stockrecords:
            return UnavailablePrice()

        try:
            # We take price from first record
            stockrecord = stockrecords[0]

            return ExcludingTaxPriceSelector(
                self.user,  # pylint: disable=no-member
                self.request,  # pylint: disable=no-member
                currency=stockrecord.price_currency,
                excl_tax=stockrecord.price_excl_tax,
                vat_rate=product.vat_rate,
            )
        except InvalidTaxRateException:
            logger.error("Tax missing for %s", product)
            return UnavailablePrice()


class UseCheapestChildProductStockRecord(UseFirstStockRecord):
    def get_product_stockrecords_queryset(self, product):
        return product.stockrecords.filter(
            price_currency__isnull=False, price_excl_tax__isnull=False
        )

    def get_product_children_queryset(self, product):
        return product.variants_ordered_by_price.filter(
            is_public=True,
            stockrecords__price_currency__isnull=False,
            stockrecords__price_excl_tax__isnull=False,
        )

    def select_stockrecord(self, product):
        pc = product.get_product_class()
        if pc.track_stock:
            return (
                self.get_product_stockrecords_queryset(product)
                .annotate(
                    db_effective_stock=db_functions.Coalesce(
                        "num_in_stock", models.Value(0)
                    )
                    - db_functions.Coalesce("num_allocated", models.Value(0))
                )
                .annotate(
                    has_stock=models.Case(
                        models.When(db_effective_stock__gt=0, then=models.Value(True)),
                        default=models.Value(False),
                        output_field=models.BooleanField(),
                    )
                )
                .order_by("-has_stock", "price_excl_tax", "-db_effective_stock")
            ).first()
        else:
            return self.get_product_stockrecords_queryset(product).first()

    def select_children_stockrecords(self, product):
        """
        Select child with best availability
        """
        pc = product.get_product_class()
        valid_children = self.get_product_children_queryset(product)

        if pc.track_stock:
            first_available_child = (
                valid_children.annotate(
                    db_effective_stock=db_functions.Coalesce(
                        "stockrecords__num_in_stock", models.Value(0)
                    )
                    - db_functions.Coalesce(
                        "stockrecords__num_allocated", models.Value(0)
                    )
                ).filter(db_effective_stock__gt=0)
            ).first()
        else:
            first_available_child = valid_children.first()

        if first_available_child is not None:
            return [
                (first_available_child, self.select_stockrecord(first_available_child))
            ]

        return []


class StockRequired(BaseStockRequired):
    """
    Availability policy mixin for use with the ``Structured`` base strategy.
    This mixin ensures that a product can only be bought if it has stock
    available (if stock is being tracked).
    """

    def availability_policy(self, product, stockrecord):
        if not stockrecord:
            return Unavailable()
        if not product.get_product_class().track_stock:
            return Available()
        else:
            return StockRequiredAvailability(product, stockrecord)



from decimal import Decimal as D
from unittest import mock

from django.test import TestCase
from django.urls import reverse

from oscar.core.loading import get_class, get_model

import ocyan.core.fender
from ocyan.core.utils import round_to_two_dec

Product = get_model("catalogue", "Product")
Basket = get_model("basket", "Basket")
Selector = get_class("partner.strategy", "Selector")

OCYAN_MOCK_CONFIG_INCL_TAX = {
    "oscar": {"use_price_incl_tax": True},
    "shipping": {
        "enable_free_shipping": False,
        "enable_weightbased_shipping": True,
        "enable_charged_shipping": True,
        "paid_shipping_first": True,
    },
}
OCYAN_MOCK_CONFIG_EXCL_TAX = {
    "oscar": {"use_price_incl_tax": False},
    "shipping": {
        "enable_free_shipping": False,
        "enable_weightbased_shipping": True,
        "enable_charged_shipping": True,
        "paid_shipping_first": True,
    },
}


@mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_EXCL_TAX)
class FourDecimalsExclTaxRoundingTest(TestCase):
    fixtures = ["taxrounding"]

    def test_price(self):
        product = Product.objects.get(pk=11544)
        strategy = Selector().strategy()
        purchase_info = strategy.fetch_for_product(product)
        self.assertEqual(purchase_info.price.excl_tax, D("0.16"))
        self.assertEqual(purchase_info.price.incl_tax, D("0.1936"))

    def test_add_to_basket(self):
        basket_add_url = reverse("basket:add", args=(11544,))
        body = {"quantity": 2}
        response = self.client.post(basket_add_url, body, follow=True)
        basket = response.context["basket"]
        self.assertEqual(basket.total_incl_tax, D("0.39"))  # 0.3872
        self.assertEqual(basket.total_excl_tax, D("0.32"))
        (line,) = basket.all_lines()

        self.assertEqual(line.unit_effective_price, D("0.1936"))
        self.assertEqual(line.unit_price_incl_tax, D("0.1936"))
        self.assertEqual(line.unit_price_excl_tax, D("0.16"))
        self.assertEqual(line.line_price_excl_tax * D("1.21"), line.line_price_incl_tax)

    def test_add_to_basket_ten(self):
        basket_add_url = reverse("basket:add", args=(11544,))
        body = {"quantity": 10}
        response = self.client.post(basket_add_url, body, follow=True)
        basket = response.context["basket"]
        self.assertEqual(basket.total_incl_tax, D("1.94"))  # 1.9360
        self.assertEqual(basket.total_excl_tax, D("1.60"))
        (line,) = basket.all_lines()

        self.assertEqual(line.unit_effective_price, D("0.1936"))
        self.assertEqual(line.unit_price_incl_tax, D("0.1936"))
        self.assertEqual(line.unit_price_excl_tax, D("0.16"))
        self.assertEqual(line.line_price_excl_tax * D("1.21"), line.line_price_incl_tax)

    def test_basket_total(self):
        self.test_add_to_basket()

        basket_add_url = reverse("basket:add", args=(1,))
        body = {"quantity": 2}
        response = self.client.post(basket_add_url, body, follow=True)
        basket = response.context["basket"]
        self.assertEqual(
            basket.total_incl_tax, D("0.94"), "The rounded sum should be 0.94"
        )
        self.assertEqual(basket.total_excl_tax, D("0.78"))
        _, line = basket.all_lines()
        self.assertEqual(line.unit_effective_price, D("0.2783"))
        self.assertEqual(line.unit_price_incl_tax, D("0.2783"))
        self.assertEqual(line.unit_price_excl_tax, D("0.23"))
        self.assertEqual(line.line_price_excl_tax * D("1.21"), line.line_price_incl_tax)


@mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_INCL_TAX)
class FourDecimalsInclTaxRoundingTest(TestCase):
    fixtures = ["taxrounding"]

    def test_price(self):
        product = Product.objects.get(pk=11544)
        strategy = Selector().strategy()
        purchase_info = strategy.fetch_for_product(product)
        self.assertEqual(
            purchase_info.price.excl_tax,
            D("0.1322"),
            "nothing should happen because it is",
        )
        self.assertEqual(purchase_info.price.incl_tax, D("0.16"))

    def test_add_to_basket(self):
        basket_add_url = reverse("basket:add", args=(11544,))
        body = {"quantity": 7}
        response = self.client.post(basket_add_url, body, follow=True)
        basket = response.context["basket"]
        self.assertEqual(basket.total_incl_tax, D("1.12"))
        self.assertEqual(basket.total_excl_tax, D("0.93"))
        self.assertEqual(basket.total_tax, D("0.19"))
        (line,) = basket.all_lines()

        self.assertEqual(line.unit_effective_price, D("0.16"))
        self.assertEqual(line.unit_price_incl_tax, D("0.16"))
        self.assertEqual(line.unit_price_excl_tax, D("0.1322"))
        self.assertEqual(
            round_to_two_dec(line.unit_price_excl_tax * D("1.21")),
            line.unit_price_incl_tax,
        )

    def test_add_to_basket_ten(self):
        basket_add_url = reverse("basket:add", args=(11544,))
        body = {"quantity": 10}
        response = self.client.post(basket_add_url, body, follow=True)
        basket = response.context["basket"]
        self.assertEqual(basket.total_incl_tax, D("1.60"))
        self.assertEqual(basket.total_excl_tax, D("1.32"))
        (line,) = basket.all_lines()

        self.assertEqual(line.unit_effective_price, D("0.16"))
        self.assertEqual(line.unit_price_incl_tax, D("0.16"))
        self.assertEqual(line.unit_price_excl_tax, D("0.1322"))
        self.assertEqual(
            round_to_two_dec(line.unit_price_excl_tax * D("1.21")),
            line.unit_price_incl_tax,
        )

    def test_basket_total(self):
        self.test_add_to_basket()

        basket_add_url = reverse("basket:add", args=(1,))
        body = {"quantity": 53}
        response = self.client.post(basket_add_url, body, follow=True)
        basket = response.context["basket"]

        self.assertEqual(basket.total_incl_tax, D("13.31"))
        self.assertEqual(basket.total_excl_tax, D("11.00"))
        _, line = basket.all_lines()
        self.assertEqual(line.unit_effective_price, D("0.23"))
        self.assertEqual(line.unit_price_incl_tax, D("0.23"))
        self.assertEqual(line.unit_price_excl_tax, D("0.1901"))
        self.assertEqual(
            round_to_two_dec(line.unit_price_excl_tax * D("1.21")),
            line.unit_price_incl_tax,
        )



# -*- coding: utf-8 -*-
from django.apps import AppConfig


class OcyanPluginOscarPartnerAppConfig(AppConfig):
    name = "ocyan.plugin.oscar_partner"
    label = "oscar_partner"

    def ready(self):
        from . import signals  # pylint: disable=unused-import, unused-variable



from unittest import mock

from django.test import TestCase

from oscar.core.loading import get_class, get_model

import ocyan.core.fender
from ocyan.plugin.oscar.fender import InvalidTaxRateException

Product = get_model("catalogue", "Product")
Selector = get_class("partner.strategy", "Selector")
tax_rate_as_fraction = get_class("taxes.taxrate", "tax_rate_as_fraction")

EXCL_TAX_SETTINGS = {"oscar": {"use_price_incl_tax": False}}


class TaxPriceSelectorTest(TestCase):
    fixtures = ["partner/auth.json", "partner/catalogue.json"]

    def test_tax_rate_as_fraction(self):
        with self.assertRaises(InvalidTaxRateException):
            tax_rate_as_fraction(None)

    def test_tax_price_selector_incl_tax(self):
        strategy = Selector().strategy()
        p = Product.objects.get()
        price_info = strategy.fetch_for_product(p)

        self.assertTrue(
            price_info.price.is_tax_known,
            "A valid product should have a known tax value",
        )
        self.assertTrue(price_info.price.exists, "A valid product should have a price")
        self.assertEqual(
            price_info.price.tax_code,
            p.vat_rate,
            "The tax code of the price should mirror the vat_rate on the product",
        )

        Product.objects.update(vat_rate=None)
        p.refresh_from_db()

        price_info = strategy.fetch_for_product(p)
        self.assertFalse(
            price_info.price.exists,
            "A product without a vat_rate should now have a price",
        )
        self.assertFalse(
            price_info.price.is_tax_known,
            "A product without vat_rate should not have a know tax value",
        )

    @mock.patch.object(ocyan.core.fender.config, "settings", EXCL_TAX_SETTINGS)
    def test_tax_price_selector_excl_tax(self):
        strategy = Selector().strategy()
        p = Product.objects.get()
        price_info = strategy.fetch_for_product(p)

        self.assertTrue(
            price_info.price.is_tax_known,
            "A valid product should have a known tax value",
        )
        self.assertTrue(price_info.price.exists, "A valid product should have a price")
        self.assertEqual(
            price_info.price.tax_code,
            p.vat_rate,
            "The tax code of the price should mirror the vat_rate on the product",
        )

        Product.objects.update(vat_rate=None)
        p.refresh_from_db()

        price_info = strategy.fetch_for_product(p)
        self.assertFalse(
            price_info.price.exists,
            "A product without a vat_rate should now have a price",
        )
        self.assertFalse(
            price_info.price.is_tax_known,
            "A product without vat_rate should not have a know tax value",
        )



from django.utils import timezone

from oscar.core.loading import get_model

from ocyan.plugin.oscar_partner.partner.constants import EU_APPLICABLE_COUNTRIES

Order = get_model("order", "Order")


def get_local_vat_order_queryset():
    year = timezone.now().year
    orders = Order.objects.filter(
        shipping_address__country__in=EU_APPLICABLE_COUNTRIES,
        date_placed__year=year,
        status__in=["paid", "complete"],
    )

    # This exists because the vat rules for 2021 only apply from july 1st until december 31st. From 2022 all orders count
    if year == 2021:
        orders = orders.filter(date_placed__month__gt=6)

    return orders



from django.core.management.base import BaseCommand

from oscar.core.loading import get_model

LocalEUVatSettings = get_model("oscar_partner", "LocalEUVatSettings")


class Command(BaseCommand):
    def handle(self, *args, **options):
        eu_settings = LocalEUVatSettings.load()
        eu_settings.vat_threshold_exceeded = False
        eu_settings.save()



# -*- coding: utf-8 -*-
from django.db import models, router
from django.db.models import F, signals
from django.db.models.functions import Coalesce, Least
from django.utils.translation import gettext_lazy as _

from oscar.apps.partner import abstract_models as oscar
from oscar.apps.partner.exceptions import InvalidStockAdjustment


class AbstractStockRecord(oscar.AbstractStockRecord):
    num_in_stock = models.IntegerField(_("Number in stock"), blank=True, null=True)

    @property
    def price_incl_tax(self):
        """
        Since oscar is about to move to specifying price without indicating
        incl or excl tax levels, we are now using the recommended way of
        letting users specify prices including tax and having the strategy
        calculate the tax price. Which is by storing the price incl tax
        in the price excl tax field.
        """
        return self.price_excl_tax

    @price_incl_tax.setter
    def price_incl_tax(self, value):
        self.price_excl_tax = value

    def allocate(self, quantity):
        """
        Record a stock allocation.
        This normally happens when a product is bought at checkout.  When the
        product is actually shipped, then we 'consume' the allocation.
        """
        # Doesn't make sense to allocate if stock tracking is off.
        if not self.can_track_allocations:
            return

        # Send the pre-save signal
        self.pre_save_signal()

        # Atomic update
        (
            self.__class__.objects.filter(pk=self.pk).update(
                num_allocated=(Coalesce(F("num_allocated"), 0) + quantity)
            )
        )

        # Make sure the current object is up-to-date
        self.refresh_from_db(fields=["num_allocated"])

        # Send the post-save signal
        self.post_save_signal()

    def consume_allocation(self, quantity):
        """
        Consume a previous allocation
        This is used when an item is shipped.  We remove the original
        allocation and adjust the number in stock accordingly
        """
        if not self.can_track_allocations:
            return
        if not self.is_allocation_consumption_possible(quantity):
            raise InvalidStockAdjustment(_("Invalid stock consumption request"))

        # send the pre save signal
        self.pre_save_signal()

        # Atomically consume allocations and stock
        (
            self.__class__.objects.filter(pk=self.pk).update(
                num_allocated=(Coalesce(F("num_allocated"), 0) - quantity),
                num_in_stock=(Coalesce(F("num_in_stock"), 0) - quantity),
            )
        )

        # Make sure current object is up-to-date
        self.refresh_from_db(fields=["num_allocated", "num_in_stock"])

        # Send the post-save signal
        self.post_save_signal()

    def cancel_allocation(self, quantity):
        if not self.can_track_allocations:
            return

        # send the pre save signal
        self.pre_save_signal()

        # Atomically consume allocations
        (
            self.__class__.objects.filter(pk=self.pk).update(
                num_allocated=Coalesce(F("num_allocated"), 0)
                - Least(Coalesce(F("num_allocated"), 0), quantity),
            )
        )

        # Make sure current object is up-to-date
        self.refresh_from_db(fields=["num_allocated"])

        # Send the post-save signal
        self.post_save_signal()

    def pre_save_signal(self):
        signals.pre_save.send(
            sender=self.__class__,
            instance=self,
            created=False,
            raw=False,
            using=router.db_for_write(self.__class__, instance=self),
        )

    def post_save_signal(self):
        signals.post_save.send(
            sender=self.__class__,
            instance=self,
            created=False,
            raw=False,
            using=router.db_for_write(self.__class__, instance=self),
        )

    class Meta(oscar.AbstractStockRecord.Meta):
        abstract = True



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ("partner", "0003_auto_20150604_1450"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="partner",
            options={
                "ordering": ("name", "code"),
                "verbose_name": "Fulfillment partner",
                "verbose_name_plural": "Fulfillment partners",
                "permissions": (("dashboard_access", "Can access dashboard"),),
            },
        ),
    ]



# Generated by Django 2.0.7 on 2018-11-15 19:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("partner", "0004_auto_20160107_1755"),
    ]

    operations = [
        migrations.AlterField(
            model_name="partner",
            name="name",
            field=models.CharField(
                blank=True, db_index=True, max_length=128, verbose_name="Name"
            ),
        ),
        migrations.AlterField(
            model_name="stockalert",
            name="date_created",
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, verbose_name="Date Created"
            ),
        ),
    ]



# -*- coding: utf-8 -*-
default_app_config = "ocyan.plugin.oscar_partner.apps.OcyanPluginOscarPartnerAppConfig"

__all__ = ["OcyanOscar3PartnerAppPluginConfig"]


def __getattr__(name):
    if name != "OcyanOscar3PartnerAppPluginConfig":
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    from .entrypoint import OcyanOscar3PartnerAppPluginConfig

    return OcyanOscar3PartnerAppPluginConfig



# Generated by Django 4.2.11 on 2024-06-03 09:54

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("oscar_partner", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="localeuvatsettings",
            name="site",
        ),
    ]



from django.core.cache import cache
from django.dispatch import receiver

from ocyan.plugin.oscar.signals import order_ready

from ocyan.plugin.oscar_partner.partner.constants import LOCAL_VAT_APPLICABLE_KEY


@receiver(order_ready)
def clear_local_vat_cache(sender, **kwargs):  # pylint: disable=unused-argument
    cache.delete(LOCAL_VAT_APPLICABLE_KEY)



# pylint: disable=W0614,W0401
from oscar.core.loading import get_class, is_model_registered

AbstractStockRecord = get_class("partner.abstract_models", "AbstractStockRecord")

###############################################################################
# This is how to override an oscar app, while keeping the overriding app      #
# overridable itself.                                                         #
# The mechanism is the same as in oscar, the only think different is the way  #
# ``__all__`` is being updated with the exports of the original oscar app.    #
###############################################################################

__all__ = []

if not is_model_registered("partner", "StockRecord"):

    class StockRecord(AbstractStockRecord):
        pass

    __all__.append("StockRecord")

# both these lines are reruired, the first one imports all objects the are not
# implemented above into the current namespace.
# the second line updates the __all__ variable of the current namespace with
# the objects from the oscar namespace.
# THIS ENSURE THAT THE SAME CAN BE DONE BY ANOTHER APP, SO THIS APP CAN BE ONCE
# MORE EXTENDED BY ANOTHER APP.

from oscar.apps.partner.models import *
from oscar.apps.partner.models import __all__ as some

__all__.extend(some)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):
    dependencies = [
        ("partner", "0002_auto_20141007_2032"),
    ]

    operations = [
        migrations.AlterField(
            model_name="partner",
            name="users",
            field=models.ManyToManyField(
                related_name="partners",
                verbose_name="Users",
                to=settings.AUTH_USER_MODEL,
                blank=True,
            ),
        ),
    ]



from ocyan.core.plugins import OcyanPluginConfig
from ocyan.core.utils import merge_dicts


class OcyanOscar3PartnerAppPluginConfig(OcyanPluginConfig):
    def contribute_to_settings(self, settings):
        return merge_dicts(
            settings,
            {
                "OCYAN_PARTNER_RELATIVE_PRICE": True,
                "OCYAN_PARTNER_EU_VAT_RULES_ACTIVATED": True,
                "MIGRATION_MODULES": {
                    "partner": "ocyan.plugin.oscar_partner.oxyan_partner_extension.migrations"
                },
            },
        )

    def contribute_to_installed_apps(self, installed_apps):
        installed_apps.ocyan.append("ocyan.plugin.oscar_partner")
        installed_apps.oscar.append(
            "ocyan.plugin.oscar_partner.partner.apps.PartnerConfig",
            replaces="oscar.apps.partner.apps.PartnerConfig",
        )
        installed_apps.oscar.register_extension(
            "ocyan.plugin.oscar_partner.oxyan_partner_extension",
            "ocyan.plugin.oscar_partner.partner",
        )
        return installed_apps



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.models.fields.autoslugfield
import oscar.models.fields
from django.conf import settings


class Migration(migrations.Migration):
    dependencies = [
        ("catalogue", "0001_initial"),
        ("address", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Partner",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "code",
                    oscar.models.fields.autoslugfield.AutoSlugField(
                        populate_from="name",
                        unique=True,
                        verbose_name="Code",
                        max_length=128,
                        editable=False,
                        blank=True,
                    ),
                ),
                (
                    "name",
                    models.CharField(max_length=128, verbose_name="Name", blank=True),
                ),
                (
                    "users",
                    models.ManyToManyField(
                        related_name="partners",
                        blank=True,
                        verbose_name="Users",
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Fulfillment partners",
                "verbose_name": "Fulfillment partner",
                "abstract": False,
                "permissions": (("dashboard_access", "Can access dashboard"),),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="PartnerAddress",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        verbose_name="Title",
                        max_length=64,
                        blank=True,
                        choices=[
                            ("Mr", "Mr"),
                            ("Miss", "Miss"),
                            ("Mrs", "Mrs"),
                            ("Ms", "Ms"),
                            ("Dr", "Dr"),
                        ],
                    ),
                ),
                (
                    "first_name",
                    models.CharField(
                        max_length=255, verbose_name="First name", blank=True
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        max_length=255, verbose_name="Last name", blank=True
                    ),
                ),
                (
                    "line1",
                    models.CharField(
                        max_length=255, verbose_name="First line of address"
                    ),
                ),
                (
                    "line2",
                    models.CharField(
                        max_length=255,
                        verbose_name="Second line of address",
                        blank=True,
                    ),
                ),
                (
                    "line3",
                    models.CharField(
                        max_length=255, verbose_name="Third line of address", blank=True
                    ),
                ),
                (
                    "line4",
                    models.CharField(max_length=255, verbose_name="City", blank=True),
                ),
                (
                    "state",
                    models.CharField(
                        max_length=255, verbose_name="State/County", blank=True
                    ),
                ),
                (
                    "postcode",
                    oscar.models.fields.UppercaseCharField(
                        max_length=64, verbose_name="Post/Zip-code", blank=True
                    ),
                ),
                (
                    "search_text",
                    models.TextField(
                        editable=False,
                        verbose_name="Search text - used only for searching addresses",
                    ),
                ),
                (
                    "country",
                    models.ForeignKey(
                        verbose_name="Country",
                        to="address.Country",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "partner",
                    models.ForeignKey(
                        verbose_name="Partner",
                        related_name="addresses",
                        to="partner.Partner",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Partner addresses",
                "verbose_name": "Partner address",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="StockAlert",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("threshold", models.PositiveIntegerField(verbose_name="Threshold")),
                (
                    "status",
                    models.CharField(
                        default="Open",
                        max_length=128,
                        verbose_name="Status",
                        choices=[("Open", "Open"), ("Closed", "Closed")],
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Date Created"
                    ),
                ),
                (
                    "date_closed",
                    models.DateTimeField(
                        blank=True, verbose_name="Date Closed", null=True
                    ),
                ),
            ],
            options={
                "ordering": ("-date_created",),
                "verbose_name_plural": "Stock alerts",
                "verbose_name": "Stock alert",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="StockRecord",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "partner_sku",
                    models.CharField(max_length=128, verbose_name="Partner SKU"),
                ),
                (
                    "price_currency",
                    models.CharField(
                        default="GBP", max_length=12, verbose_name="Currency"
                    ),
                ),
                (
                    "price_excl_tax",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        blank=True,
                        verbose_name="Price (excl. tax)",
                        null=True,
                    ),
                ),
                (
                    "price_retail",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        blank=True,
                        verbose_name="Price (retail)",
                        null=True,
                    ),
                ),
                (
                    "cost_price",
                    models.DecimalField(
                        max_digits=12,
                        decimal_places=2,
                        blank=True,
                        verbose_name="Cost Price",
                        null=True,
                    ),
                ),
                (
                    "num_in_stock",
                    models.PositiveIntegerField(
                        blank=True, verbose_name="Number in stock", null=True
                    ),
                ),
                (
                    "num_allocated",
                    models.IntegerField(
                        blank=True, verbose_name="Number allocated", null=True
                    ),
                ),
                (
                    "low_stock_threshold",
                    models.PositiveIntegerField(
                        blank=True, verbose_name="Low Stock Threshold", null=True
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Date created"
                    ),
                ),
                (
                    "date_updated",
                    models.DateTimeField(
                        auto_now=True, db_index=True, verbose_name="Date updated"
                    ),
                ),
                (
                    "partner",
                    models.ForeignKey(
                        verbose_name="Partner",
                        related_name="stockrecords",
                        to="partner.Partner",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        verbose_name="Product",
                        related_name="stockrecords",
                        to="catalogue.Product",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Stock records",
                "verbose_name": "Stock record",
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name="stockrecord",
            unique_together=set([("partner", "partner_sku")]),
        ),
        migrations.AddField(
            model_name="stockalert",
            name="stockrecord",
            field=models.ForeignKey(
                verbose_name="Stock Record",
                related_name="alerts",
                to="partner.StockRecord",
                on_delete=models.CASCADE,
            ),
            preserve_default=True,
        ),
    ]



import decimal
from collections import namedtuple

from django import forms
from django.conf import settings
from django.core.cache import cache
from django.db.models import Sum
from django.db.models.functions import Coalesce
from django.utils.translation import gettext as _

from oscar.core.decorators import deprecated
from oscar.core.loading import get_class, get_model

from ocyan.core.utils import (
    round_to_two_dec as deprecated_round_to_two_dec,
)
from ocyan.core.utils import (
    to_money as deprecated_to_money,
)
from ocyan.plugin.oscar.taxes.constants import COUNTRY_TAX_RATES
from ocyan.plugin.oscar.taxes.utils import (
    exclusive_tax_component as deprecated_exclusive_tax_component,
)
from ocyan.plugin.oscar.taxes.utils import (
    get_country_specific_tax_code,
)
from ocyan.plugin.oscar.taxes.utils import (
    inclusive_tax_component as deprecated_inclusive_tax_component,
)
from ocyan.plugin.oscar.taxes.utils import (
    tax_from_price_incl_tax as deprecated_tax_from_price_incl_tax,
)

from .constants import (
    EU_APPLICABLE_COUNTRIES,
    EU_LOCAL_VAT_TRESHOLD,
    LOCAL_VAT_APPLICABLE_KEY,
)

decode_option_value = get_class("basket.option", "option_display_value")
tax_rate_as_fraction = get_class("taxes.taxrate", "tax_rate_as_fraction")
get_local_vat_order_queryset = get_class(
    "partner.queryset", "get_local_vat_order_queryset"
)

round_to_two_dec = deprecated(deprecated_round_to_two_dec)
to_money = deprecated(deprecated_to_money)
exclusive_tax_component = deprecated(deprecated_exclusive_tax_component)
inclusive_tax_component = deprecated(deprecated_inclusive_tax_component)
tax_from_price_incl_tax = deprecated(deprecated_tax_from_price_incl_tax)

LocalEUVatSettings = get_model("oscar_partner", "LocalEUVatSettings")

TaxInfo = namedtuple("TaxInfo", ["percentage", "tax_code"])


def get_vat_country_from_request(request):
    if request is not None:
        return request.META.get(
            "shipping_country", settings.OSCAR_VAT_COUNTRY_OF_ORIGIN_CODE
        )

    return settings.OSCAR_VAT_COUNTRY_OF_ORIGIN_CODE


def is_local_vat_applicable(country):
    if not settings.OCYAN_PARTNER_EU_VAT_RULES_ACTIVATED:
        return False

    if country in EU_APPLICABLE_COUNTRIES:
        local_vat_applicable = cache.get(LOCAL_VAT_APPLICABLE_KEY)
        if local_vat_applicable is None:
            eu_settings = LocalEUVatSettings.load()
            if eu_settings.vat_threshold_exceeded:
                cache.set(LOCAL_VAT_APPLICABLE_KEY, True, 60 * 60)
                local_vat_applicable = True
            else:
                orders = get_local_vat_order_queryset()

                total_sum = orders.aggregate(
                    total_sum=Coalesce(Sum("total_incl_tax"), decimal.Decimal("0.00"))
                )["total_sum"]

                if total_sum > decimal.Decimal(EU_LOCAL_VAT_TRESHOLD):
                    cache.set(LOCAL_VAT_APPLICABLE_KEY, True, 60 * 60)
                    local_vat_applicable = True

        return local_vat_applicable

    return False


@deprecated
def get_vat_rate_for_country(
    vat_rate, country, is_shipping_charge, basket=None, user=None
):
    return get_vat_info_for_country(
        vat_rate, country, is_shipping_charge, basket=basket, user=user
    )


# pylint: disable=unused-argument
def get_vat_info_for_country(
    vat_rate, country, is_shipping_charge, basket=None, user=None
):
    if is_local_vat_applicable(country):
        if is_shipping_charge and "shipping" in COUNTRY_TAX_RATES[country]:
            vat_rate = "shipping"

        return TaxInfo(
            tax_rate_as_fraction(vat_rate, country=country),
            get_country_specific_tax_code(vat_rate, country),
        )
    else:
        return TaxInfo(tax_rate_as_fraction(vat_rate), vat_rate)


def get_name_and_value(option_value):
    """
    Depending if the option comes from the posted data
    of from the database, it could be a dict or an object.
    This function extracts the name and the value regardless of which it is.
    The value will be treated as if it was a json value.

    >>> from argparse import Namespace
    >>> option = Namespace(code="ruuu")
    >>> attr_opt = Namespace(option=option, value="1")
    >>> dict_opt = {"option": option, "value": "[1, 2, 3]"}
    >>> get_name_and_value(attr_opt)
    ('ruuu', '1')
    >>> get_name_and_value(dict_opt)
    ('ruuu', '[1, 2, 3]')
    """
    try:
        option = option_value["option"]
        value = option_value["value"]
    except TypeError:
        option = option_value.option
        value = option_value.value

    return option.code, decode_option_value(value)


class ParsedOptions(dict):
    """
    Access product options like they where structured in a normal way as
    attributes::

        a = ParsedOptions(POST['options])
        a["shirt_text"]

    or::

        a = ParsedOptions(line.options)
        a["shirt_text"]

    The obvious usecase is to parse the options passed to the strategy in::

        def fetch_for_product(self, product, options=None, stockrecord=None):
            pass
    """

    def __init__(self, option_list):
        if option_list is not None:
            super().__init__([get_name_and_value(option) for option in option_list])
        else:
            super().__init__()

    def validate(self, *names):
        errors = []
        for name in names:
            if name not in self:
                errors.append(name)
        if errors:
            raise forms.ValidationError(_("Option %(errors)s missing") % locals())
